# Copyright 2018 AstroLab Software
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from py4j.java_gateway import JavaObject
from py4j.java_collections import JavaList

from pyspark3d import load_from_jvm
from pyspark3d import get_spark_context

from pyspark import RDD
from pyspark.mllib.common import _java2py

def scala2java(scala_list: JavaObject) -> JavaList:
    """
    Convert a Scala list coming from the JVM into a JavaList:
    scala.collection.immutable.List -> java.util.ArrayList

    Parameters
    ----------
    scala_list : JavaObject
        Scala list (scala.collection.immutable.List)

    Returns
    ----------
    java_list : JavaList
        Java list (java.util.ArrayList)

    Examples
    ----------
    Instantiate a Point3D
    >>> from pyspark3d.geometryObjects import Point3D
    >>> p = Point3D(0.0, 0.0, 0.0, True)

    Grab the coordinates
    >>> coord_scala = p.getCoordinate()
    >>> assert(type(coord_scala) == JavaObject)

    `coord_scala` is a Scala List. Convert it to Java List
    >>> coord_java = scala2java(coord_scala)
    >>> assert(type(coord_java) == JavaList)
    """
    # Not very powerful check... As many things other than Scala list
    # are JavaObject...
    msg = """
    converters.scala2java expects a JavaObject not {}.
    """.format(type(scala_list))
    assert(type(scala_list) == JavaObject), msg

    convpath = "scala.collection.JavaConverters.seqAsJavaListConverter"
    converter = load_from_jvm(convpath)
    return converter(scala_list).asJava()

def java2python(java_list: JavaList) -> list:
    """
    Convert a Java list into a python list:
    java.util.ArrayList -> list

    Parameters
    ----------
    java_list : JavaList
        Java list (java.util.ArrayList)

    Returns
    ----------
    python_list : list
        Python list

    Examples
    ----------
    Instantiate a Point3D
    >>> from pyspark3d.geometryObjects import Point3D
    >>> p = Point3D(0.0, 0.0, 0.0, True)

    Grab the coordinates
    >>> coord_scala = p.getCoordinate()
    >>> assert(type(coord_scala) == JavaObject)

    `coord_scala` is a Scala List. Convert it to Java List
    >>> coord_java = scala2java(coord_scala)
    >>> assert(type(coord_java) == JavaList)

    Convert then the Java List into a python list
    >>> coord_python = java2python(coord_java)
    >>> assert(type(coord_python) == list)

    Other example. Instantiate directly a JavaList in python
    >>> java_list = sc._jvm.java.util.ArrayList()
    >>> java_list.add(0.0) # return a boolean
    True

    >>> python_list = java2python(java_list)
    >>> assert(python_list[0] == 0.0)
    """
    msg = """
    converters.java2python expects a JavaList not {}.
    """.format(type(java_list))
    assert(type(java_list) == JavaList), msg

    return list(java_list)

def scala2python(scala_list: JavaObject) -> list:
    """
    Convert a Scala list coming from the JVM into a python list:
    scala.collection.immutable.List -> list

    Parameters
    ----------
    scala_list : JavaObject
        Scala list (scala.collection.immutable.List)

    Returns
    ----------
    python_list : list
        Python list

    Examples
    ----------
    Instantiate a Point3D
    >>> from pyspark3d.geometryObjects import Point3D
    >>> p = Point3D(0.0, 0.0, 0.0, True)

    Grab the coordinates
    >>> coord_scala = p.getCoordinate()
    >>> assert(type(coord_scala) == JavaObject)

    `coord_scala` is a Scala List. Convert it to python List
    >>> coord_python = scala2python(coord_scala)
    >>> assert(type(coord_python) == list)
    """
    # Not very powerful check... As many things other than Scala list
    # are JavaObject...
    msg = """
    converters.scala2python expects a JavaObject not {}.
    """.format(type(scala_list))
    assert(type(scala_list) == JavaObject), msg

    return java2python(scala2java(scala_list))


if __name__ == "__main__":
    """
    Run the doctest using

    python converters.py

    If the tests are OK, the script should exit gracefuly, otherwise the
    failure(s) will be printed out.
    """
    import os
    import sys
    import doctest
    import numpy as np

    from pyspark import SparkContext
    from pyspark3d import pyspark3d_conf
    from pyspark3d import load_user_conf
    from pyspark3d import get_spark_session

    # Activate the SparkContext for the test suite
    dic = load_user_conf()
    conf = pyspark3d_conf("local[*]", "test", dic)
    sc = SparkContext.getOrCreate(conf=conf)

    # Load the spark3D JAR+deps, and initialise the spark session.
    # In a pyspark shell, you do not need this.
    spark = get_spark_session(dicconf=dic)

    # Numpy introduced non-backward compatible change from v1.14.
    if np.__version__ >= "1.14.0":
        np.set_printoptions(legacy="1.13")

    # Run the test suite
    failure_count, test_count = doctest.testmod(extraglobs={"spark": spark})
    sys.exit(failure_count)



# Copyright 2018 AstroLab Software
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from pyspark.mllib.common import _java2py
from pyspark.sql import DataFrame

from pyspark3d import load_from_jvm
from pyspark3d import get_spark_context

def knn(df: DataFrame, p: list, k: int, coordSys: str, unique: bool):
    """ Finds the K nearest neighbors of the query object.

    The naive implementation here searches through all the the objects in
    the DataFrame to get the KNN. The nearness of the objects here
    is decided on the basis of the distance between their centers.

    Parameters
    ----------
    df : DataFrame
        Input Dataframe. Must have 3 columns corresponding to the
        coordinate (x, y, z) if cartesian or (r, theta, phi) f spherical.
    p : list of float
        Targeted point for which we want neighbors.
    k : int
        Number of neighbours
    coordSys : str
        Coordinate system: spherical or cartesian
    unique : bool
        Boolean. If true, returns only distinct objects. Default is false.

    Returns
    --------
    out : DataFrame
        DataFrame with the coordinates of the k neighbours found.

    Examples
    --------
    >>> df = spark.read.format("fits")\
        .option("hdu", 1)\
        .load("../src/test/resources/cartesian_points.fits")

    Get the 100 closest neighbours around the point [0.2, 0.2, 0.2]
    >>> K = 100
    >>> target = [0.2, 0.2, 0.2]
    >>> unique = False
    >>> neighbours = knn(df.select("x", "y", "z"), target, K, "spherical", unique)

    >>> print(neighbours.count())
    100

    You can add back the metadata
    >>> neighboursWithMeta = df.join(neighbours, ["x", "y", "z"], "left_semi")
    """
    prefix = "com.astrolabsoftware.spark3d"
    scalapath = "{}.Queries.KNN".format(prefix)
    scalaclass = load_from_jvm(scalapath)

    # # To convert python List to Scala Map
    convpath = "{}.python.PythonClassTag.javaListtoscalaList".format(prefix)
    conv = load_from_jvm(convpath)

    out = _java2py(get_spark_context() ,scalaclass(df._jdf, conv(p), k, coordSys, unique))

    return out

def windowquery(df: DataFrame, windowtype: str, windowcoord: int, coordSys: str):
    """ Perform window query, that is match between DF elements and
    a user-defined window (point, sphere, shell, box).

    If windowtype =
        - point: windowcoord = [x, y, z]
        - sphere: windowcoord = [x, y, z, R]
        - shell: windowcoord = [x, y, z, Rin, Rout]
        - box: windowcoord = [x1, y1, z1, x2, y2, z2, x3, y3, z3]
    Use [x, y, z] for cartesian or [r, theta, phi] for spherical.
    Note that box only accepts cartesian coordinates.

    Parameters
    ----------
    df : DataFrame
        Input Dataframe. Must have 3 columns corresponding to the
        coordinate (x, y, z) if cartesian or (r, theta, phi) f spherical.
    windowtype : str
        point, shell, sphere, or box.
    windowcoord : list of float
        List of Doubles. The coordinates of the window (see doc above).
    coordSys : str
        Coordinate system: spherical or cartesian

    Returns
    --------
    out : DataFrame
        DataFrame with the coordinates of the objects found in the window

    Examples
    --------
    >>> df = spark.read.format("csv")\
        .option("inferSchema", True)\
        .option("header", True)\
        .load("../src/test/resources/cartesian_spheres_manual.csv")

    Point-like window
    >>> windowtype = "point"
    >>> windowcoord = [1.0, 1.0, 1.0]
    >>> env = windowquery(df.select("x", "y", "z"), windowtype, windowcoord, "cartesian")

    >>> print(env.count())
    2

    You can add back the metadata
    >>> envWithMeta = df.join(env, ["x", "y", "z"], "left_semi")

    Sphere-like window
    >>> windowtype = "sphere"
    >>> windowcoord = [1.0, 1.0, 1.0, 2.0]
    >>> env = windowquery(df.select("x", "y", "z"), windowtype, windowcoord, "cartesian")

    >>> print(env.count())
    3

    Shell-like window
    >>> windowtype = "shell"
    >>> windowcoord = [1.0, 1.0, 1.0, 0.0, 2.0]
    >>> env = windowquery(df.select("x", "y", "z"), windowtype, windowcoord, "cartesian")

    >>> print(env.count())
    3

    Box-like window
    >>> windowtype = "box"
    >>> windowcoord = [2.0, 0.0, 0.0, 0.0, 2.0, 0.0, 0.0, 0.0, 2.0]
    >>> env = windowquery(df.select("x", "y", "z"), windowtype, windowcoord, "cartesian")

    >>> print(env.count())
    2
    """
    prefix = "com.astrolabsoftware.spark3d"
    scalapath = "{}.Queries.windowQuery".format(prefix)
    scalaclass = load_from_jvm(scalapath)

    # # To convert python List to Scala Map
    convpath = "{}.python.PythonClassTag.javaListtoscalaList".format(prefix)
    conv = load_from_jvm(convpath)

    out = _java2py(get_spark_context() ,scalaclass(df._jdf, windowtype, conv(windowcoord), coordSys))

    return out


if __name__ == "__main__":
    """
    Run the doctest using

    python queries.py

    If the tests are OK, the script should exit gracefuly, otherwise the
    failure(s) will be printed out.
    """
    import sys
    import doctest
    import numpy as np

    from pyspark import SparkContext
    from pyspark3d import pyspark3d_conf
    from pyspark3d import load_user_conf
    from pyspark3d import get_spark_session
    # Activate the SparkContext for the test suite
    dic = load_user_conf()
    conf = pyspark3d_conf("local", "test", dic)
    sc = SparkContext.getOrCreate(conf=conf)

    # Load the spark3D JAR+deps, and initialise the spark session.
    # In a pyspark shell, you do not need this.
    spark = get_spark_session(dicconf=dic)

    # Numpy introduced non-backward compatible change from v1.14.
    if np.__version__ >= "1.14.0":
        np.set_printoptions(legacy="1.13")

    # Run the test suite
    failure_count, test_count = doctest.testmod()
    sys.exit(failure_count)



# Copyright 2018 AstroLab Software
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from pyspark.sql import DataFrame
from pyspark.mllib.common import _java2py

from pyspark3d import load_from_jvm
from pyspark3d import get_spark_context

def checkLoadBalancing(df: DataFrame, kind: str="frac", numberOfElements: int=-1):
    """
    DataFrame containing the weight of each partition.
    You can choose between outputing the size (number of rows) of each partition
    or the fractional size (%) to the total number of rows.
    size of the dataset (in percent). This is useful to check whether the
    load is correctly balanced.

    Parameters
    ----------
    df : DataFrame
        Input DataFrame
    kind : str
        print the load balancing in terms of fractional size (kind="frac")
        or number of rows per partition (kind="size"). Default is "frac".
    numberOfElements : int
        (optional). Total number of elements in the DataFrame.
        Only needed if you choose to output fractional sizes (kind="frac").
        If not provided (i.e. default value of -1) and kind="frac",
        it will be computed (count).

    Returns
    ----------
    dfout : DataFrame containing the weight of each partition.

    Examples
    ----------
    Load data
    >>> df = spark.read.format("fits")\
        .option("hdu", 1)\
        .load("../src/test/resources/astro_obs.fits")

    Fake repartitioning in 10 equal sized partitions
    >>> df = df.repartition(10)

    Compute the load balancing %
    >>> df_load = checkLoadBalancing(df, kind="frac")

    Note that this is a DataFrame, so you can use df.show()
    Here we will check that the total is indeed 100%
    >>> val = df_load.select("Load (%)").collect()
    >>> assert(int(sum([i[0] for i in val])) == 100)

    Same using number of rows instead of fractional contribution
    >>> df_load = checkLoadBalancing(df, kind="size")
    >>> val = df_load.select("Load (#Rows)").collect()

    >>> assert(int(sum([i[0] for i in val])) == df.count())
    """
    prefix = "com.astrolabsoftware.spark3d"
    scalapath = "{}.Checkers.checkLoadBalancing".format(prefix)
    scalaclass = load_from_jvm(scalapath)

    dfout = _java2py(
        get_spark_context(),
        scalaclass(df._jdf, kind, numberOfElements))

    return dfout


if __name__ == "__main__":
    """
    Run the doctest using

    python checkers.py

    If the tests are OK, the script should exit gracefuly, otherwise the
    failure(s) will be printed out.
    """
    import sys
    import doctest
    import numpy as np

    from pyspark import SparkContext
    from pyspark3d import pyspark3d_conf
    from pyspark3d import load_user_conf
    from pyspark3d import get_spark_session
    # Activate the SparkContext for the test suite
    dic = load_user_conf()
    conf = pyspark3d_conf("local", "test", dic)
    sc = SparkContext.getOrCreate(conf=conf)

    # Load the spark3D JAR+deps, and initialise the spark session.
    # In a pyspark shell, you do not need this.
    spark = get_spark_session(dicconf=dic)

    # Numpy introduced non-backward compatible change from v1.14.
    if np.__version__ >= "1.14.0":
        np.set_printoptions(legacy="1.13")

    # Run the test suite
    failure_count, test_count = doctest.testmod()
    sys.exit(failure_count)



# Copyright 2018 AstroLab Software
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from pyspark3d import load_from_jvm

from py4j.java_gateway import JavaObject

def Point3D(x: float, y: float, z: float, isSpherical: bool) -> JavaObject:
    """
    Binding around Point3D.scala. For full description,
    see `$spark3d/src/main/scala/com/spark3d/geometryObjects/Point3D.scala`.

    By default, the input coordinates are supposed euclidean,
    that is (x, y, z). The user can also work with spherical input coordinates
    (x=r, y=theta, z=phi) by setting the argument isSpherical=true.

    Parameters
    ----------
    x : float
        Input X coordinate in Euclidean space, and R in spherical space.
    y : float
        Input Y coordinate in Euclidean space, and THETA in spherical space.
    z : float
        Input Z coordinate in Euclidean space, and PHI in spherical space.
    isSpherical : bool
        If true, it assumes that the coordinates of the Point3D
        are (r, theta, phi). Otherwise, it assumes cartesian
        coordinates (x, y, z).

    Returns
    ----------
    p3d : Point3D instance
        An instance of the class Point3D.

    Example
    ----------
    Instantiate a point with spherical coordinates (r, theta, phi)
    >>> p3d = Point3D(1.0, np.pi, 0.0, True)

    The returned type is JavaObject (Point3D instance)
    >>> print(type(p3d))
    <class 'py4j.java_gateway.JavaObject'>

    You can then call the method associated, for example
    >>> p3d.getVolume()
    0.0

    Return the point coordinates
    >>> p3d = Point3D(1.0, 1.0, 0.0, False)
    >>> p3d.getCoordinatePython()
    [1.0, 1.0, 0.0]

    It will be a JavaList by default
    >>> coord = p3d.getCoordinatePython()
    >>> print(type(coord))
    <class 'py4j.java_collections.JavaList'>

    Make it a python list
    >>> coord_python = list(coord)
    >>> print(type(coord_python))
    <class 'list'>

    [Astro] Convert the (theta, phi) in Healpix pixel index:
    >>> p3d = Point3D(1.0, np.pi, 0.0, True) # (z, theta, phi)
    >>> p3d.toHealpix(2048, True)
    50331644

    To see all the available methods:
    >>> print(sorted(p3d.__dir__())) # doctest: +NORMALIZE_WHITESPACE
    ['center', 'distanceTo', 'equals', 'getClass', 'getCoordinate',
    'getCoordinatePython', 'getEnvelope', 'getHash', 'getVolume',
    'hasCenterCloseTo', 'hashCode', 'intersects', 'isEqual', 'isSpherical',
    'notify', 'notifyAll', 'toHealpix', 'toHealpix$default$2', 'toString',
    'wait', 'x', 'y', 'z']
    """
    scalapath = "com.astrolabsoftware.spark3d.geometryObjects.Point3D"
    p3d = load_from_jvm(scalapath)

    return p3d(x, y, z, isSpherical)

def ShellEnvelope(*args) -> JavaObject:
    """
    Binding around ShellEnvelope.scala. For full description, see
    `$spark3d/src/main/scala/com/spark3d/geometryObjects/ShellEnvelope.scala`

    The Scala version makes use of several constructors (i.e. with different
    kinds of argument). In order to mimick this within a single routine, we
    abstract the arguments of the constructor using the iterable `*args`.
    There are then 5 possibilities to instantiate a `ShellEnvelope`:

    Case 1: Defined with a center coordinates, inner and outer radius.
        args = [x: Double, y: Double, z: Double,
            isSpherical: Boolean, innerRadius: Double, outerRadius: Double]
    Case 2: Defined with a center coordinates, and a radius (= a sphere).
        args = [x: Double, y: Double, z: Double,
            isSpherical: Boolean, radius: Double]
    Case 3: Defined with a Point3D, and a radius (= a sphere).
        args = [p: Point3D(...), radius: Double]
    Case 4: from another ShellEnvelope
        args = [shell: ShellEnvelope(...)]
    Case 5: Null envelope
        args = []

    Returns
    ----------
    shell : ShellEnvelope instance
        An instance of the class ShellEnvelope. Throw an error if
        the iterable in the constructor is not understood.

    Example
    ----------
    >>> from pyspark3d.geometryObjects import Point3D

    Case 1: Defined with a center coordinates (cart), inner and outer radius.
    >>> shell_case1 = ShellEnvelope(0.0, 1.0, 1.0, False, 0.5, 1.0)
    >>> assert("ShellEnvelope" in shell_case1.__str__())

    Case 2: Defined with a center coordinates, and a radius (= a sphere).
    >>> shell_case2 = ShellEnvelope(0.0, 0.0, 0.0, False, 1.0)
    >>> print(round(shell_case2.getArea(), 1))
    12.6

    Case 3: Defined with a Point3D, and a radius (= a sphere).
    >>> origin = Point3D(0.0, 0.0, 0.0, False)
    >>> shell_case3 = ShellEnvelope(origin, 1.0)
    >>> print(shell_case3.intersects(origin))
    True

    Case 4: From another ShellEnvelope
    >>> shell_case4 = ShellEnvelope(shell_case3)
    >>> print(shell_case4.isEqual(shell_case3))
    True

    Case 5: The null shell
    >>> shell_case5 = ShellEnvelope()
    >>> print(shell_case5.isNull())
    True

    To see all the available methods:
    >>> print(sorted(shell_case1.__dir__())) # doctest: +NORMALIZE_WHITESPACE
    ['center', 'contains', 'equals', 'expandBy', 'expandInnerRadius',
    'expandOuterRadius', 'expandToInclude', 'getArea', 'getClass',
    'getEnvelope', 'getHash', 'hasCenterCloseTo', 'hashCode', 'innerRadius',
    'innerRadius_$eq', 'intersects', 'intersectsShell', 'isEqual', 'isNull',
    'isPointInShell', 'notify', 'notifyAll', 'outerRadius', 'outerRadius_$eq',
    'setToNull', 'toHealpix', 'toHealpix$default$2', 'toString', 'wait']

    """
    warning = """
        There are 5 possibilities to instantiate a `ShellEnvelope`:

        Case 1: Defined with a center coordinates, inner and outer radius.
            args = [x: Double, y: Double, z: Double,
                isSpherical: Boolean, innerRadius: Double, outerRadius: Double]
        Case 2: Defined with a center coordinates, and a radius (= a sphere).
            args = [x: Double, y: Double, z: Double,
                isSpherical: Boolean, radius: Double]
        Case 3: Defined with a Point3D, and a radius (= a sphere).
            args = [p: Point3D(...), radius: Double]
        Case 4: from another ShellEnvelope
            args = [shell: ShellEnvelope(...)]
        Case 5: Null envelope
            args = []
    """
    scalapath = "com.astrolabsoftware.spark3d.geometryObjects.ShellEnvelope"
    shell = load_from_jvm(scalapath)

    # case 6
    if len(args) == 0:
        return shell()

    # Case 5
    elif len(args) == 1:
        cond_shell = "ShellEnvelope" in args[0].__str__()

        msg = """
        You are trying to instantiate a ShellEnvelope with 1 argument which is
        not a ShellEnvelope.

        {}
        """.format(warning)

        assert(cond_shell), msg

        return shell(args[0])

    # Case 3
    elif len(args) == 2:
        msg = """
        You are trying to instantiate a ShellEnvelope with 2 arguments
        which are not a Point3D (center) and a float (radius).

        {}
        """.format(warning)

        assert("Point3D" in args[0].__str__()), msg
        assert(type(args[1]) is float or type(args[1]) is int), msg

        return shell(args[0], args[1])

    # Case 2
    elif len(args) == 5:
        msg = """
        You are trying to instantiate a ShellEnvelope with 5 arguments
        but there is one or several type mismatch.

        {}
        """.format(warning)

        assert(type(args[0]) == int or type(args[0]) == float), msg
        assert(type(args[1]) == int or type(args[1]) == float), msg
        assert(type(args[2]) == int or type(args[2]) == float), msg
        assert(type(args[3]) == bool), msg
        assert(type(args[4]) == int or type(args[4]) == float), msg

        return shell(args[0], args[1], args[2], args[3], args[4])
    # Case 1
    elif len(args) == 6:
        msg = """
        You are trying to instantiate a ShellEnvelope with 6 arguments
        but there is one or several type mismatch.

        {}
        """.format(warning)

        assert(type(args[0]) == int or type(args[0]) == float), msg
        assert(type(args[1]) == int or type(args[1]) == float), msg
        assert(type(args[2]) == int or type(args[2]) == float), msg
        assert(type(args[3]) == bool), msg
        assert(type(args[4]) == int or type(args[4]) == float), msg
        assert(type(args[5]) == int or type(args[5]) == float), msg

        return shell(args[0], args[1], args[2], args[3], args[4], args[5])
    else:
        msg = """
        Constructor not understood.

        {}
        """.format(warning)
        assert(False), msg

def BoxEnvelope(*args) -> JavaObject:
    """
    Binding around BoxEnvelope.scala. For full description,
    see `$spark3d/src/main/scala/com/spark3d/geometryObjects/BoxEnvelope.scala`

    The Scala version makes use of several constructors (i.e. with different
    kinds of argument). In order to mimick this within a single routine, we
    abstract the arguments of the constructor using the iterable `*args`.
    There are then 5 possibilities to instantiate a `BoxEnvelope`:

    Case 1: from coordinates
        args = [x1: float, x2: float, y1: float, y2: float,
            z1: float, z2: float]
    Case 2: from a single Point3D (i.e. the box is a Point3D)
        args = [p: Point3D(...)]
    Case 3: from three Point3D
        args = [p1: Point3D(...), p2: Point3D(...), p3: Point3D(...)]
    Case 4: from another BoxEnvelope
        args = [b: BoxEnvelope(...)]
    Case 5: Null envelope
        args = []

    Coordinates of input Point3D MUST be cartesian.

    Returns
    ----------
    box : BoxEnvelope instance
        An instance of the class BoxEnvelope. Throw an error if the
        iterable in the constructor is not understood.

    Example
    ----------
    >>> from pyspark3d.geometryObjects import Point3D

    Case 1: Cube from coordinates
    >>> box_case1 = BoxEnvelope(0.0, 1.0, 0.0, 1.0, 0.0, 1.0)
    >>> print(box_case1.__str__())
    Env[0.0 : 1.0, 0.0 : 1.0, 0.0 : 1.0, ]

    Case 2: Zero volume
    >>> p3d = Point3D(0.0, 0.0, 0.0, False)
    >>> box_case2 = BoxEnvelope(p3d)
    >>> print(box_case2.getVolume())
    0.0

    Case 3: Cube from 3 Point3D
    >>> p3d_1 = Point3D(0.0, 1.0, 0.0, False)
    >>> p3d_2 = Point3D(0.1, 1.0, 0.0, False)
    >>> p3d_3 = Point3D(1.0, -1.0, 1.0, False)
    >>> origin = Point3D(0.0, 0.0, 0.0, False)
    >>> box_case3 = BoxEnvelope(p3d_1, p3d_2, p3d_3)
    >>> print(box_case3.contains(origin))
    True

    Case 4: From another envelope
    >>> box_case4 = BoxEnvelope(box_case3)
    >>> print(box_case4.isEqual(box_case3))
    True

    Case 5: The null cube
    >>> box_case5 = BoxEnvelope()
    >>> print(box_case5.isNull())
    True

    To see all the available methods:
    >>> print(sorted(box_case1.__dir__())) # doctest: +NORMALIZE_WHITESPACE
    ['apply', 'center', 'contains', 'covers', 'distance', 'equals', 'expandBy',
    'expandOutwards', 'expandToInclude', 'getClass', 'getEnvelope', 'getHash',
    'getVolume', 'getXLength', 'getYLength', 'getZLength', 'hasCenterCloseTo',
    'hashCode', 'indexID', 'indexID_$eq', 'intersection', 'intersects',
    'intersectsBox', 'intersectsRegion', 'isEqual', 'isNull', 'maxExtent',
    'maxX', 'maxX_$eq', 'maxY', 'maxY_$eq', 'maxZ', 'maxZ_$eq', 'minExtent',
    'minX', 'minX_$eq', 'minY', 'minY_$eq', 'minZ', 'minZ_$eq', 'notify',
    'notifyAll', 'setToNull', 'toHealpix', 'toHealpix$default$2', 'toString',
    'translate', 'wait']

    """
    warning = """
        There are 5 possibilities to instantiate a `BoxEnvelope`:

        Case 1: from coordinates
            args = [x1: float, x2: float, y1: float, y2: float,
                z1: float, z2: float]
        Case 2: from a single Point3D (i.e. the box is a Point3D)
            args = [p: Point3D(...)]
        Case 3: from three Point3D
            args = [p1: Point3D(...), p2: Point3D(...), p3: Point3D(...)]
        Case 4: from another BoxEnvelope
            args = [b: BoxEnvelope(...)]
        Case 5: Null envelope
            args = []
    """
    scalapath = "com.astrolabsoftware.spark3d.geometryObjects.BoxEnvelope"
    box = load_from_jvm(scalapath)

    # case 6
    if len(args) == 0:
        return box()

    # Case 2 or 4
    elif len(args) == 1:
        cond_p3d = "Point3D" in args[0].__str__()
        cond_box = "Env" in args[0].__str__()

        msg = """
        You are trying to instantiate a BoxEnvelope with 1 argument which is
        neither a Point3D nor a BoxEnvelope.

        {}
        """.format(warning)

        assert(cond_p3d or cond_box), msg

        return box(args[0])

    # Case 3
    elif len(args) == 3:
        msg = """
        You are trying to instantiate a BoxEnvelope with 3 arguments and one
        at least is not a Point3D.

        {}
        """.format(warning)

        for arg in args:
            assert("Point3D" in arg.__str__()), msg

        return box(args[0], args[1], args[2])

    # Case 3
    elif len(args) == 6:
        msg = """
        You are trying to instantiate a BoxEnvelope with 3 arguments and one
        at least is not a Point3D.

        {}
        """.format(warning)

        for arg in args:
            assert(type(arg) == int or type(arg) == float), msg

        return box(args[0], args[1], args[2], args[3], args[4], args[5])
    else:
        msg = """
        Constructor not understood.

        {}
        """.format(warning)
        assert(False), msg


if __name__ == "__main__":
    """
    Run the doctest using

    python geometryObjects.py

    If the tests are OK, the script should exit gracefuly, otherwise the
    failure(s) will be printed out.
    """
    import sys
    import doctest
    import numpy as np

    from pyspark import SparkContext
    from pyspark3d import pyspark3d_conf
    from pyspark3d import load_user_conf
    # Activate the SparkContext for the test suite
    dic = load_user_conf()
    conf = pyspark3d_conf("local", "test", dic)
    sc = SparkContext.getOrCreate(conf=conf)

    # Numpy introduced non-backward compatible change from v1.14.
    if np.__version__ >= "1.14.0":
        np.set_printoptions(legacy="1.13")

    # Run the test suite
    failure_count, test_count = doctest.testmod()
    sys.exit(failure_count)



# Copyright 2018 AstroLab Software
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from pyspark.sql import DataFrame
from pyspark.mllib.common import _java2py

from pyspark3d import load_from_jvm
from pyspark3d import get_spark_context

from typing import Dict

def prePartition(df: DataFrame, options: Dict={"": ""}, numPartitions: int=-1):
    """Add a DataFrame column describing the partitioning.

    This method allows to use a custom partitioner (SpatialPartitioner).
    Note that no data movement (shuffle) is performed yet here, as we just
    describe how the repartitioning should be done.
    Use `partitionBy` to trigger it.

    `options` must contain four entries:
       - gridtype: the type of repartitioning. Available: current (no repartitioning), onion, octree.
       - geometry: geometry of objects: points, spheres, or boxes
       - coordSys: coordinate system: spherical or cartesian
       - colnames: comma-separated names of the spatial coordinates. For points,
            must be "x,y,z" or "r,theta,phi". For spheres, must be "x,y,z,R" or
            "r,theta,phi,R".

    Parameters
    ----------
    df : DataFrame
        Input DataFrame
    options : Dictionary of Strings
        Dictionary containing metadata (see above).
    numPartitions : int
        (optional) The number of partitions wanted. -1 by default,
        i.e. the number of partitions of the input DataFrame.

    Returns
    ----------
    dfout : DataFrame
        Input DataFrame plus an additional column `partition_id`.

    Examples
    ----------
    Load data
    >>> df = spark.read.format("fits")\
        .option("hdu", 1)\
        .load("../src/test/resources/astro_obs.fits")

    Specify options
    >>> options = {
    ...     "geometry": "points",
    ...     "colnames": "Z_COSMO,RA,DEC",
    ...     "coordSys": "spherical",
    ...     "gridtype": "onion"}

    Add a column containing the partitioning (Onion)
    >>> df_colid = prePartition(df, options, 10)
    >>> print(df_colid.select("partition_id").distinct().count())
    10

    Note that you can also return the current partitioning:
    >>> options = {
    ...     "geometry": "points",
    ...     "colnames": "Z_COSMO,RA,DEC",
    ...     "coordSys": "spherical",
    ...     "gridtype": "current"}
    >>> df_colid = prePartition(df, options)
    >>> assert(df_colid.select("partition_id").distinct().count() == df.rdd.getNumPartitions())
    """
    prefix = "com.astrolabsoftware.spark3d"
    scalapath = "{}.Repartitioning.prePartition".format(prefix)
    scalaclass = load_from_jvm(scalapath)

    # To convert python dic to Scala Map
    convpath = "{}.python.PythonClassTag.javaHashMaptoscalaMap".format(prefix)
    conv = load_from_jvm(convpath)

    dfout = _java2py(
        get_spark_context(),
        scalaclass(df._jdf, conv(options), numPartitions))

    return dfout

def repartitionByCol(df: DataFrame, colname: str, preLabeled: bool, numPartitions: int=-1):
    """Repartition a DataFrame according to a column containing partition ID.

    Note this is not re-ordering elements, but making new partitions with
    objects having the same partition ID defined by one of the
    DataFrame column (i.e. shuffling).

    Parameters
    ----------
    df : DataFrame
        Input DataFrame.
    colname : str
        Column name describing the repartitioning. Typically Ints.
    preLabeled : bool
        True means the column containing the partition ID contains
        already numbers from 0 to `numPartitions - 1`. false otherwise.
        Note that in the latter, the execution time will be longer as
        we need to map column values to partition ID.
    numPartitions : int
        (optional )Number of partitions. If not provided the code will
        guess the number of partitions by counting the number of distinct
        elements of the repartitioning column.
        As it can be costly, you can provide manually this information.

    Returns
    ---------
    dfout : DataFrame
        Repartitioned input DataFrame.

    Examples
    ---------
    Load data
    >>> df = spark.read.format("fits")\
        .option("hdu", 1)\
        .load("../src/test/resources/astro_obs.fits")

    Specify options
    >>> options = {
    ...     "geometry": "points",
    ...     "colnames": "Z_COSMO,RA,DEC",
    ...     "coordSys": "spherical",
    ...     "gridtype": "onion"}

    Add a column containing the partitioning (Onion)
    >>> df_colid = prePartition(df, options, 10)
    >>> print(df_colid.select("partition_id").distinct().count())
    10

    Trigger the repartitioning
    >>> df_repart = repartitionByCol(df_colid, "partition_id", True, 10)
    >>> def mapLen(part): yield len([*part])
    >>> df_repart.rdd.mapPartitions(mapLen).take(1)[0]
    2104
    """
    prefix = "com.astrolabsoftware.spark3d"
    scalapath = "{}.Repartitioning.repartitionByCol".format(prefix)
    scalaclass = load_from_jvm(scalapath)

    dfout = _java2py(
        get_spark_context(),
        scalaclass(df._jdf, colname, preLabeled, numPartitions))

    return dfout


if __name__ == "__main__":
    """
    Run the doctest using

    python repartitioning.py

    If the tests are OK, the script should exit gracefuly, otherwise the
    failure(s) will be printed out.
    """
    import sys
    import doctest
    import numpy as np

    from pyspark import SparkContext
    from pyspark3d import pyspark3d_conf
    from pyspark3d import load_user_conf
    from pyspark3d import get_spark_session
    # Activate the SparkContext for the test suite
    dic = load_user_conf()
    conf = pyspark3d_conf("local", "test", dic)
    sc = SparkContext.getOrCreate(conf=conf)

    # Load the spark3D JAR+deps, and initialise the spark session.
    # In a pyspark shell, you do not need this.
    spark = get_spark_session(dicconf=dic)

    # Numpy introduced non-backward compatible change from v1.14.
    if np.__version__ >= "1.14.0":
        np.set_printoptions(legacy="1.13")

    # Run the test suite
    failure_count, test_count = doctest.testmod()
    sys.exit(failure_count)



# Copyright 2018 AstroLab Software
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from pyspark import SparkConf
from pyspark import SparkContext
from pyspark.sql import SparkSession

import inspect
from typing import Any, List, Dict

from pyspark3d.pyspark3d_conf import extra_jars
from pyspark3d.pyspark3d_conf import extra_packages
from pyspark3d.pyspark3d_conf import log_level

from pyspark3d.version import __version__

ROOT_JVM = "_gateway.jvm"
MSG = """WARNING: You should not use {} outside the test suite or in a regular
pyspark session, or in batch mode.
"""
global TEST
TEST = False

def get_spark_context() -> SparkContext:
    """
    Return the current SparkContext.
    Raise a RuntimeError if spark hasn't been initialized.

    Returns
    ---------
    sparkContext : SparkContext instance
        The active sparkContext

    Examples
    ---------
    >>> pysc = get_spark_context()
    >>> print(type(pysc))
    <class 'pyspark.context.SparkContext'>
    """
    if SparkContext._active_spark_context:
        return SparkContext._active_spark_context
    else:
        raise RuntimeError("SparkContext must be initialized")

def rec_load(obj: Any, mod: List[str], count: int=0) -> Any:
    """
    Load recursively JavaPackages and JavaClasses residing inside the JVM:
    python world -> gateway -> JVM -> my_scala_packages

    There is no guarantee your package exist though!
    See the example below for the syntax.

    Parameters
    ----------
    obj : SparkContext instance or Any
        Initial call must take a SparkContext (`pyspark.context.SparkContext`).
        Then obj will represent subsequently:
            - `py4j.java_gateway.JavaGateway`
            - `py4j.java_gateway.JVMView`
            - `py4j.java_gateway.JavaPackage`
            - `py4j.java_gateway.JavaPackage`
            - ...
    mod : str
        List of packages from the SparkContext to your class in the JVM
    count : int, optional
        Counter used for the recursion. Must be 0 at the initial call.

    Returns
    ----------
    obj : Any
        obj is an instance of a JVM object and will represent subsequently:
            - `py4j.java_gateway.JavaGateway`
            - `py4j.java_gateway.JVMView`
            - `py4j.java_gateway.JavaPackage`
            - `py4j.java_gateway.JavaPackage`
            - ...

    Example
    ----------
    >>> pysc = get_spark_context()
    >>> mod = "_gateway.jvm.com.astrolabsoftware.spark3d.geometryObjects"
    >>> jvm_obj = rec_load(pysc, mod.split("."))
    >>> print(type(jvm_obj))
    <class 'py4j.java_gateway.JavaPackage'>
    """
    if count == len(mod):
        return obj
    else:
        return rec_load(getattr(obj, mod[count]), mod, count+1)

def load_from_jvm(scala_package: str) -> Any:
    """
    Load a Scala package (instance of a JVM object) from Python.

    Note: If the object returned by `load_from_jvm` is a JavaPackage while you
    were expecting a JavaClass, that means it has not been found in the JVM.
    Most likely you forgot to include the spark3D JAR in the classpath.

    Parameters
    ----------
    scala_package : str
        Scala package path as if you were importing it in Scala.

    Returns
    ----------
    jvm_obj : Any
        instance of a JVM object: `py4j.java_gateway.JavaPackage` or
        `py4j.java_gateway.JavaClass`.

    Example
    ----------
    >>> pysc = get_spark_context()
    >>> sp = "com.astrolabsoftware.spark3d.geometryObjects.Point3D"
    >>> jvm_obj = load_from_jvm(sp)
    >>> print(type(jvm_obj))
    <class 'py4j.java_gateway.JavaClass'>
    """
    path = ROOT_JVM + "." + scala_package
    packages = path.split(".")

    # Get the SparkContext
    pysc = get_spark_context()

    # Load from Scala to Python
    jvm_obj = rec_load(pysc, packages)

    return jvm_obj

def get_spark_session(
        master: str="local[*]",
        appname: str="test",
        dicconf={}) -> SparkSession:
    """
    Return a SparkSession with extra configuration parameters.

    Note
    ----------
    this routine is mainly used in 2 contexts:
      - Running the test suite
      - Using spark3D within a standard ipython or notebook session
    In those two cases, you need to load the JAR+deps within the session.

    In a regular pyspark session, or in batch mode, you should not use it.
    Instead, link spark3D when starting the session:

    pyspark --jars /path/to/spar3djar --master ...

    spark-submit --jars /path/to/spar3djar --master ...

    If you installed pyspark3d via pip, the JAR is released with the
    python source files. To find its location:
    `python -c "import pyspark3d; print(pyspark3d.__file__)"`

    Parameters
    ----------
    master : str
        Execution mode: local[*], spark://..., yarn, etc
        If you are already in a Spark Session (pyspark shell), this won't have
        any effect.
    appname : str
        Name for the application
    dicconf : dictionary
        Key/value to pass to the SparkConf. Typically location of JARS,
        Maven coordinates, etc..

    Returns
    ----------
    spark : SparkSession
        The spark session

    Examples
    ----------
    >>> dicconf = load_user_conf()
    >>> spark = get_spark_session("local[*]", "test", dicconf)
    >>> print(type(spark))
    <class 'pyspark.sql.session.SparkSession'>
    """
    if not TEST:
        print(MSG.format(inspect.currentframe().f_code.co_name))

    # Grab the user conf
    conf = pyspark3d_conf(master, appname, dicconf)

    # Instantiate a spark session
    spark = SparkSession\
        .builder\
        .appName("test")\
        .config(conf=conf)\
        .getOrCreate()

    return spark

def load_user_conf(sparkSession: SparkSession=None) -> Dict:
    """
    Load pre-defined user Spark configuration stored in pyspark3d_conf.py
    to be passed to the SparkConf.
    If a spark Session is already running, you can pass it to transfer the
    conf parameters.

    Note
    ----------
    this routine is mainly used in 2 contexts:
      - Running the test suite
      - Using spark3D within a standard ipython or notebook session
    In those two cases, you need to load the JAR+deps within the session.

    In a regular pyspark session, or in batch mode, you should not use it.
    Instead, link spark3D when starting the session:

    pyspark --jars /path/to/spar3djar --master ...

    spark-submit --jars /path/to/spar3djar --master ...

    If you installed pyspark3d via pip, the JAR is released with the
    python source files. To find its location:
    `python -c "import pyspark3d; print(pyspark3d.__file__)"`

    Parameters
    ----------
    sparkSession : SparkSession, optional
        If set, the returned dictionary contains the
        parameters from this spark Session. Default is None.

    Returns
    ---------
    dic : dictionary
        Dictionary with extra arguments to be passed to the SparkConf

    Examples
    ---------
    From scratch
    >>> dic = load_user_conf()
    >>> assert("spark.jars" in dic)

    From a previous session
    >>> spark = SparkSession.builder.appName("toto").getOrCreate()
    >>> dic = load_user_conf(spark)
    >>> print(dic["spark.app.name"])
    toto
    >>> assert("spark.jars" in dic)
    """
    if not TEST:
        print(MSG.format(inspect.currentframe().f_code.co_name))

    extra_jars_with_commas = ",".join(extra_jars)
    extra_packages_with_commas = ",".join(extra_packages)

    if sparkSession is not None:
        oldconf = sparkSession.sparkContext.getConf().getAll()
        dic = {}
        for item in oldconf:
            k, v = item
            if k == "spark.jars":
                v += "," + extra_jars_with_commas
            if k == "spark.jars.packages":
                v += "," + extra_packages_with_commas
            dic[k] = v
    else:
        dic = {
            "spark.jars": extra_jars_with_commas,
            "spark.jars.packages": extra_packages_with_commas
            }

    return dic

def pyspark3d_conf(
        master: str, AppName: str, confdic: Dict={}) -> SparkConf:
    """
    Set the configuration for running pyspark3d.
    In case you have a doubt about a missing package, just run:
    `conf.toDebugString().split("\n")`
    to see what is registered in the conf.

    Note
    ----------
    this routine is mainly used in 2 contexts:
      - Running the test suite
      - Using spark3D within a standard ipython or notebook session
    In those two cases, you need to load the JAR+deps within the session.

    In a regular pyspark session, or in batch mode, you should not use it.
    Instead, link spark3D when starting the session:

    pyspark --jars /path/to/spar3djar --master ...

    spark-submit --jars /path/to/spar3djar --master ...

    If you installed pyspark3d via pip, the JAR is released with the
    python source files. To find its location:
    `python -c "import pyspark3d; print(pyspark3d.__file__)"`

    Parameters
    ----------
    master : str
        The master URL to connect to, such as "local" to run
        locally with one thread, "local[4]" to run locally with 4 cores, or
        "spark://master:7077" to run on a Spark standalone cluster, or
        "yarn" to run on a YARN cluster.
    AppName : str
        The name for the application.
    confdic : Dict, optional
        Additional key/value to be passed to the configuration.
        Typically, this is the place where you will set the path to
        external JARS.

    Returns
    ----------
    conf : SparkConf instance

    Examples
    ----------
    >>> dic = load_user_conf()
    >>> conf = pyspark3d_conf("local[*]", "myTest", dic)
    >>> conf.get("spark.master")
    'local[*]'
    """
    if not TEST:
        print(MSG.format(inspect.currentframe().f_code.co_name))

    conf = SparkConf()
    conf.setMaster(master)
    conf.setAppName(AppName)
    for k, v in confdic.items():
        conf.set(key=k, value=v)

    return conf

def set_spark_log_level(log_level_manual=None):
    """
    Set the level of log in Spark.

    Parameters
    ----------
    log_level_manual : String, optional
        Level of log wanted: INFO, WARN, ERROR, OFF, etc.
        By default, this is read from pyspark3d_conf.py

    Example
    ----------
    >>> pysc = get_spark_context()

    # should be verbose or whatever Spark default has been set
    >>> rdd_verb = pysc.parallelize([1, 2, 3, 4]).collect()

    # Default is OFF
    >>> set_spark_log_level()
    >>> rdd_silent = pysc.parallelize([1, 2, 3, 4]).collect()

    # Explicitly force to be silent
    >>> set_spark_log_level("OFF")
    >>> rdd_silent = pysc.parallelize([1, 2, 3, 4]).collect()
    """
    ## Get the logger
    pysc = get_spark_context()
    logger = pysc._jvm.org.apache.log4j

    ## Set the level
    if log_level_manual is None:
        level = getattr(logger.Level, log_level, "OFF")
    else:
        level = getattr(logger.Level, log_level_manual, "INFO")

    logger.LogManager.getLogger("org").setLevel(level)
    logger.LogManager.getLogger("akka").setLevel(level)


if __name__ == "__main__":
    """
    Run the doctest using

    python __init__.py

    If the tests are OK, the script should exit gracefuly, otherwise the
    failure(s) will be printed out.
    """
    import sys
    import doctest
    import numpy as np

    # Switch to test mode
    TEST = True

    # Activate the SparkContext for the test suite
    dic = load_user_conf()
    conf = pyspark3d_conf("local", "test", dic)
    sc = SparkContext.getOrCreate(conf=conf)

    # Numpy introduced non-backward compatible change from v1.14.
    if np.__version__ >= "1.14.0":
        np.set_printoptions(legacy="1.13")

    # Run the test suite
    failure_count, test_count = doctest.testmod()
    sys.exit(failure_count)



# Copyright 2018 AstroLab Software
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
from pathlib import Path
from pyspark3d.version import __version__, __scala_version__

# In principle, users do not need to load all this stuff below.
# It is mainly used in 2 contexts:
#   - Running the test suite
#   - Using spark3D within a standard ipython or notebook session
# In those two cases, you need to load the JAR+deps within the session.

# For local tests
path_to_conf = Path().cwd().as_uri()

# This is the place where paths and environment variables should be defined
# Note that this is only for the driver (loading JARS, conf, etc)

# spark3D version
version = __version__

# Scala version used to compile spark3D
scala_version = __scala_version__

# Verbosity for Spark
log_level = "WARN"

# External JARS to be added to both driver and executors
# Should contain the FAT JAR of spark3D.
here = os.path.abspath(os.path.dirname(__file__))

# If the package has been installed with pip, the JAR
# is released with the py files
from_pip = os.path.join(here, "spark3D-assembly-{}.jar".format(version))

# If the package is built from sources, then the JAR
# is in the spark3D target/scala-{version} folder.
from_source = os.path.join(
    here, "../target/scala-{}/spark3D-assembly-{}.jar".format(
        scala_version, version))

if os.path.isfile(from_pip):
    extra_jar = from_pip
elif os.path.isfile(from_source):
    extra_jar = from_source
else:
    extra_jar = ""

extra_jars = [extra_jar]

# External packages specified using their Maven coordinates
extra_packages = [
    "com.github.astrolabsoftware:spark-fits_2.11:0.7.1",
]



# Copyright 2018 AstroLab Software
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os

import numpy as np

from typing import Iterator, Any, Callable, Generator

from py4j.java_gateway import JavaObject

from pyspark3d import load_from_jvm
from pyspark3d.converters import scala2python

from pyspark.mllib.common import _java2py
from pyspark import RDD

class CollapseFunctions():
    """ Set of functions to create a representation of the data.
    The idea is to reduce the size of the dataset while keeping desired
    features (e.g. clustering).

    We give as an example two simple examples dealing with the clustering:
    - mean (1-means)
    - k-means

    The reduction is done at the level of the Spark partition (Iterator).

    """

    def mean(self, partition: Iterator) -> Generator:
        """Compute the centroid of the partition.
        It can be viewed as a k-means for k=1.

        Parameters
        ----------
        partition : Iterator
            Iterator over the elements of the Spark partition.

        Returns
        -------
        Generator of (list, int)
            Yield tuple with the centroid and the total number of points
            in the partition. If the partition is empty, return (None, 0).

        Examples
        -------
        List of coordinates (can be 2D, 3D, ..., nD)
        >>> mylist = [[1., 2.], [3., 4.], [5., 6.], [7., 8.], [9., 10.]]

        Considering only 1 partition
        >>> cf = CollapseFunctions()
        >>> myit = iter(mylist)
        >>> list(cf.mean(myit))
        [(array([ 5.,  6.]), 5)]

        Distribute over 2 partitions
        >>> rdd = sc.parallelize(mylist, 2)

        Compute the centroid for each partition
        >>> data = rdd.mapPartitions(
        ...     lambda partition: cf.mean(partition)).collect()
        >>> print(data)
        [(array([ 2.,  3.]), 2), (array([ 7.,  8.]), 3)]

        If there are empty partitions, it returns None for the empty ones.
        >>> rdd = sc.parallelize(mylist, 6)
        >>> cf = CollapseFunctions()
        >>> data = rdd.mapPartitions(
        ...     lambda partition: cf.mean(partition)).collect()
        >>> centroids = [c[0] for c in data]
        >>> print(centroids) # doctest: +NORMALIZE_WHITESPACE
        [None, array([ 1.,  2.]), array([ 3.,  4.]), array([ 5.,  6.]),
        array([ 7.,  8.]), array([  9.,  10.])]

        """
        import numpy as np

        # Unwrap the iterator
        xyz = [*partition]
        size = len(xyz)

        # Compute the centroid only if the partition is not empty
        if size > 0:
            mean = np.mean(xyz, axis=0)
        else:
            mean = None

        yield (mean, size)

    def kmeans(self, partition: Iterator, k: int=1) -> Generator:
        """Performs k-means on a the elements of the partition.
        The case k=1 corresponds to the function `mean` above.

        Parameters
        ----------
        partition : Iterator
            Iterator over the elements of the Spark partition.
        k : int, optional
            The number of centroids to generate. Default is 1 (mean).

        Returns
        -------
        Generator of (list, int)
            The list of centroids, and the total number of points in the
            partition. If the partition is empty, return ([None], 0).

        Examples
        -------
        List of coordinates (can be 2D, 3D, ..., nD)
        >>> mylist = [
        ...     [1., 2., 4.], [3., 4., 1.], [5., 6., 5.],
        ...     [9., 10., 7.], [1., 2., 7.], [3., 4., 6.],
        ...     [5., 6., 9.], [7., 8., 6.], [9., 10., 10.]]

        Considering only 1 partition
        >>> cf = CollapseFunctions()
        >>> myit = iter(mylist[0: 2])
        >>> list(cf.kmeans(myit, 1))
        [(array([[ 2. ,  3. ,  2.5]]), 2)]

        Distribute over 2 partitions
        >>> rdd = sc.parallelize(mylist, 2)

        Compute the centroid for each partition
        >>> data = rdd.mapPartitions(
        ...     lambda partition: cf.kmeans(partition, 1)).collect()
        >>> print(data) # doctest: +NORMALIZE_WHITESPACE
        [(array([[ 4.5 ,  5.5 ,  4.25]]), 4), (array([[ 5. ,  6. ,  7.6]]), 5)]

        If there are empty partitions, it returns None for the empty ones.
        >>> rdd = sc.parallelize(mylist, 10)
        >>> cf = CollapseFunctions()
        >>> data = rdd.mapPartitions(
        ...     lambda partition: cf.kmeans(partition, 1)).collect()
        >>> centroids = [c[0] for c in data]
        >>> print(centroids) # doctest: +NORMALIZE_WHITESPACE
        [[None], array([[ 1.,  2.,  4.]]), array([[ 3.,  4.,  1.]]),
        array([[ 5.,  6.,  5.]]), array([[  9.,  10.,   7.]]),
        array([[ 1.,  2.,  7.]]), array([[ 3.,  4.,  6.]]),
        array([[ 5.,  6.,  9.]]), array([[ 7.,  8.,  6.]]),
        array([[  9.,  10.,  10.]])]

        """
        from scipy.cluster.vq import kmeans

        # Unwrap the iterator
        xyz = [*partition]
        size = len(xyz)

        if len(xyz) > k - 1:
            centers, distortion = kmeans(xyz, k)
        else:
            centers = [None]

        yield (centers, size)


def collapse_rdd_data(
        rdd: RDD, collapse_function: Callable[[Iterator, Any], Generator],
        *args: Any) -> RDD:
    """Apply a collapse function to reduce the size of the data
    set. The function is applied for each partition (mapPartitions).

    Parameters
    ----------
    rdd : RDD[list[float]]
        RDD of list of float. Each list is the coordinate of a point (x, y, z).
    collapse_function : function
        collapse function to reduce the size of the data set. See
        `CollapseFunctions` for more information.
    args: Any
        Any arguments that have to be passed to `collapse_function`.
        Must be comma-separated.

    Returns
    -------
    RDD
        RDD whose elements are the result of the collapse function for
        each partition.

    Examples
    -------
    List of coordinates (can be 2D, 3D, ..., nD)
    >>> mylist = [
    ...     [1., 2., 4.], [3., 4., 1.], [5., 6., 5.],
    ...     [9., 10., 7.], [1., 2., 7.], [3., 4., 6.],
    ...     [5., 6., 9.], [7., 8., 6.], [9., 10., 10.]]

    Distribute over 2 partitions
    >>> rdd = sc.parallelize(mylist, 2)

    Compute the centroid for each partition
    >>> cf = CollapseFunctions()
    >>> data = collapse_rdd_data(rdd, cf.kmeans, 1).collect()
    >>> print(data) # doctest: +NORMALIZE_WHITESPACE
    [(array([[ 4.5 ,  5.5 ,  4.25]]), 4), (array([[ 5. ,  6. ,  7.6]]), 5)]

    """
    return rdd.mapPartitions(
        lambda partition: collapse_function(partition, *args))

def scatter3d_mpl(x: list, y: list, z: list, radius: list=None, label: str = None, axIn = None, **kwargs):
    """3D scatter plot from matplotlib.
    Invoke show() or save the figure to get the result.

    Parameters
    ----------
    x: list of float
        X coordinate
    y: list of float
        Y coordinate
    z: list of float
        Z coordinate
    radius: int/float or list of int/float, optional
        If given, the size of the markers. Can be a single number
        of a list of sizes (of the same length as the coordinates)
    label : str
        Add a label for the legend. Default is None.
    axIn : Axes3D
        If provided, overplot. Default is None (new figure).
    kwargs : Dictionary
        Additional key/value for the plot (color, etc...)

    """
    if axIn is None:
        import pylab as pl
        from mpl_toolkits.mplot3d import Axes3D
        fig = pl.figure()
        ax = Axes3D(fig)
    else:
        ax = axIn

    # Size of the centroids
    if radius is None:
        rad = 10.
    else:
        assert len(radius) == len(x), "Wrong size!"
        rad = radius

    ax.scatter(x, y, z, s=rad, label=label, **kwargs)

    ax.set_xlabel("X")
    ax.set_ylabel("Y")
    ax.set_zlabel("Z")

    return ax

def sph2cart(r: float, theta: float, phi: float) -> (float, float, float):
    """ Conversion from spherical to cartesian

    Parameters
    ----------
    r : float
        Radial distance (>= 0)
    theta : float
        Polar angle in radian (0 <= theta <= np.pi)
    r : float
        Azimuthal angle in radian (0 <= phi <= 2 * np.pi)

    Returns
    ----------
    x : float
        Cartesian X in units of r
    y : float
        Cartesian Y in units of r
    z : float
        Cartesian Z in units of r

    Examples
    ----------
    >>> p = sph2cart(1.0, 0.0, 0.0)
    >>> print(p)
    (0.0, 0.0, 1.0)
    """
    x = r * np.sin(theta) * np.cos(phi)
    y = r * np.sin(theta) * np.sin(phi)
    z = r * np.cos(theta)
    return x, y, z

def cart2sph(x, y, z):
    """ Conversion from cartesian to spherical

    Parameters
    ----------
    x : float
        Cartesian X
    y : float
        Cartesian Y
    z : float
        Cartesian Z

    Returns
    ----------
    r : float
        Radial distance (>= 0) in the units of (X, Y, Z)
    theta : float
        Polar angle in radian (0 <= theta <= np.pi)
    r : float
        Azimuthal angle in radian (0 <= phi <= 2 * np.pi)

    Examples
    ----------
    >>> p = cart2sph(0.0, 0.0, 1.0)
    >>> print(p)
    (1.0, 0.0, 0.0)
    """
    r = np.sqrt(x*x + y*y + z*z)
    theta = np.arctan2(np.sqrt(x*x + y*y), z)
    phi = np.arctan2(y, x)
    return r, theta, phi

if __name__ == "__main__":
    """
    Run the doctest using

    python visualisation.py

    If the tests are OK, the script should exit gracefuly, otherwise the
    failure(s) will be printed out.
    """
    import sys
    import doctest
    import numpy as np

    from pyspark import SparkContext
    from pyspark3d import pyspark3d_conf
    from pyspark3d import load_user_conf

    # Activate the SparkContext for the test suite
    dic = load_user_conf()
    conf = pyspark3d_conf("local", "test", dic)
    sc = SparkContext.getOrCreate(conf=conf)

    # Numpy introduced non-backward compatible change from v1.14.
    if np.__version__ >= "1.14.0":
        np.set_printoptions(legacy="1.13")

    # Run the test suite
    failure_count, test_count = doctest.testmod()
    sys.exit(failure_count)



#!/usr/bin/env python
# Copyright 2018 AstroLab Software
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -*- coding: utf-8 -*-

import os
import subprocess

from setuptools import setup
from distutils.command.build import build
from distutils.command.clean import clean
from distutils.command.sdist import sdist
from distutils.spawn import find_executable

with open("README.md", "r") as fh:
    long_description = fh.read()

requirements = ["numpy>=1.14", "pyspark", "scipy"]
setup_requirements = ["wheel"]
test_requirements = ["coverage>=4.2", "coveralls"]

# Read the versions
exec(compile(
    open("pyspark3d/version.py").read(),
    "pyspark3d/version.py",
    "exec"))

VERSION = __version__
SCALA_VERSION = __scala_version__
SCALA_VERSION_ALL = __scala_version_all__

ASSEMBLY_JAR = \
    "target/scala-2.11/spark3D-assembly-{}.jar".format(
        __version__)
# Move the JAR inside the package
ASSEMBLY_JAR_S = \
    "pyspark3d/spark3D-assembly-{}.jar".format(
        __version__)


class jar_build(build):
    """ Class to handle spark3D JAR while installing pyspark3d """
    def run(self):
        """
        Override distutils.command.build.
        Compile the companion library and produce a FAT jar.
        """
        if find_executable('sbt') is None:
            raise EnvironmentError("""
            The executable "sbt" cannot be found.
            Please install the "sbt" tool to build the companion jar file.
            """)

        build.run(self)
        subprocess.check_call(
            "sbt 'set test in assembly := {{}}' ++{} assembly".format(
                SCALA_VERSION_ALL), shell=True)

        subprocess.check_call(
            "cp {} {}".format(
                ASSEMBLY_JAR, ASSEMBLY_JAR_S), shell=True)


class jar_clean(clean):
    """ Extends distutils.command.clean """
    def run(self):
        """
        Cleans the scala targets from the system.
        """
        clean.run(self)
        subprocess.check_call('sbt clean', shell=True)


class my_sdist(sdist):
    """ Extends distutils.command.sdist """
    def initialize_options(self, *args, **kwargs):
        """
        During installation, open the MANIFEST file
        and insert the path to the spark3D JAR required
        to run pyspark3d.
        """
        here = os.path.dirname(os.path.abspath(__file__))
        filename = os.path.join(here, "MANIFEST.in")
        with open(filename, 'w') as f:
            incld = "include {}\n"
            f.write(incld.format(ASSEMBLY_JAR_S))
        return super().initialize_options(*args, **kwargs)


setup(
    name='pyspark3d',
    version=VERSION,
    description="Spark extension for processing large-scale 3D data sets",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="AstroLab Software",
    author_email='peloton@lal.in2p3.fr',
    url='https://github.com/astrolabsoftware/spark3d',
    packages=['pyspark3d'],
    install_requires=requirements,
    license="Apache License, Version 2.0",
    zip_safe=False,
    keywords=['spark', 'spark3d', 'scala', 'python', 'py4j'],
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
    ],
    tests_require=test_requirements,
    cmdclass={
        'build': jar_build,
        'clean': jar_clean,
        'sdist': my_sdist,
    },
    include_package_data=True,
    setup_requires=setup_requirements
)
